/******* Modélisation de l’évolution d’une population de Lapins *********
                 --> TP4 de Simulation | S1 ZZ2 F2
------------------------------------------------------------------------
	       Code écrit par Anaël Courjaud en Février 2023
		Dernière modification le 05/02/2023 par Anaël Courjaud
------------------------------------------------------------------------*/


#include "general.hpp"
#include "mersenneTwister.hpp"
#include "Lapin.hpp"
 
/* Réponse à la question N° 1. C'est une suite de Fibonacci
long simRabbitBasicEvolution(int nbrMois)
{
    long nbrCouplesLapinsGenPrecedente = 1; // Un couple de jeunes lapins
    long nbrCouplesLapinsGenActuelle = 1; // Un couple de lapins matures
    long nbrCouplesLapinsGenSuivante = nbrCouplesLapinsGenActuelle; // au cas où nbrMois = 0

    for(int i = 0; i < nbrMois; i++) 
    {
        nbrCouplesLapinsGenSuivante = nbrCouplesLapinsGenActuelle + nbrCouplesLapinsGenPrecedente; // Il existe déjà nbrCouplesLapinsGenActuelle couples, à cela on y ajoute les couples produits par la génération précédente (ils sont tous devenus matures entre-temps)
        
        // On met à jour les compteurs (= on avance d'un mois)
        nbrCouplesLapinsGenPrecedente = nbrCouplesLapinsGenActuelle;
        nbrCouplesLapinsGenActuelle = nbrCouplesLapinsGenSuivante;
    }

    return nbrCouplesLapinsGenActuelle;
}
*/

int main()
{
    time_t begin = time( NULL ); // On enregistre l'heure de lancement du programme.
    cout << "\n";

    // Initialisation du Mersenne Twister
    unsigned long init[4]={0x123, 0x234, 0x345, 0x456}, length=4;
    init_by_array(init, length);

    
    double tableOfT[] = { // Utile pour le calcul de l'intervalle de confiance 
        12.706, 4.303, 3.182, 2.776, 2.571,
        2.447, 2.365, 2.308, 2.262, 2.228,
        2.201, 2.179, 2.160, 2.145, 2.131, 
        2.120, 2.110, 2.101, 2.093, 2.086,
        2.080, 2.074, 2.069, 2.064, 2.060, 
        2.056, 2.052, 2.048, 2.045, 2.042, 
        2.040, 2.038, 2.036, 2.034, 2.032,
        2.030, 2.028, 2.026, 2.024, 2.021,
        2.020, 2.018, 2.016, 2.014, 2.012,
        2.010, 2.008, 2.006, 2.004, 2.002,
        2.000, 2.000, 2.000, 2.000, 2.000, 
        2.000, 2.000, 2.000, 2.000, 2.000, 
        2.000, 2.000, 2.000, 2.000, 2.000, 
        2.000, 2.000, 2.000, 2.000, 2.000, 
        2.000, 2.000, 2.000, 2.000, 2.000, 
        2.000, 2.000, 2.000, 2.000, 2.000, 
        2.000, 2.000, 2.000, 2.000, 2.000, 
        2.000, 2.000, 2.000, 2.000, 2.000, 
        2.000, 2.000, 2.000, 2.000, 2.000, 
        2.000, 2.000, 2.000, 2.000, 2.000, 
        2.000, 2.000, 2.000, 2.000, 2.000, 
        2.000, 2.000, 2.000, 2.000, 2.000, 
        1.998, 1.996, 1.994, 1.992, 1.990,
        1.988, 1.986, 1.984, 1.982, 1.980,
        1.960
    };

    result_t resultatsExecution[NBREXECUTIONSPOURSTATISTIQUEFIABLE];

    for(int exec = 0; exec < NBREXECUTIONSPOURSTATISTIQUEFIABLE; exec++)
    {
        list<Lapin> listeLapins;
        list<list<Lapin> :: iterator> listeLapinsMorts;
        int nbrLapinsMorts = 0;

        for(int i = 0; i < NBRBEBESLAPINSINITIAUX; i++) // on commence par la naissance de 30 lapins
        {
            Lapin lapin1;
            listeLapins.push_back(lapin1);
        }

        for(int i = 0; i < NBRMOISEXPERIENECE; i++) // période de 10 mois
        {
            //cout << "début du mois " << (i + 1) << ", il y a " << listeLapins.size() << " lapins vivants et " << nbrLapinsMorts << " lapins sont morts en tout\n";
            for(list<Lapin> :: iterator itLapin = listeLapins.begin(); itLapin != listeLapins.end(); itLapin++)
            {
                if((*itLapin).survie())
                {
                    if((*itLapin).fairePetits())
                    {
                        int nombrePetits = uniform_int(MINPETITS, MAXPETITS);
                        //cout << nombrePetits << " petits\n";
                        for(int petit = 0; petit < nombrePetits; petit++)
                        {
                            Lapin nouveauLapin;
                            listeLapins.push_front(nouveauLapin);
                        }
                    }
                }
                else
                {
                    listeLapinsMorts.push_back(itLapin);
                }
            }

            for(list<list<Lapin> :: iterator> :: iterator itLapinMort = listeLapinsMorts.begin(); itLapinMort != listeLapinsMorts.end(); itLapinMort++)
            {
                listeLapins.erase(*itLapinMort);
                nbrLapinsMorts++;
            }
            listeLapinsMorts.clear();
        }
        resultatsExecution[exec].nbrLapinsMorts = nbrLapinsMorts;
        resultatsExecution[exec].nbrLapinsVivants = listeLapins.size();
        resultatsExecution[exec].sommeNaissanceDeLapins = nbrLapinsMorts + resultatsExecution[exec].nbrLapinsVivants;
    }

    ///////////////// Interprétation des résultas ////////////////////////

    // Calcul de la moyenne et de la proportion d'exécutions où l'espéce ne réussit pas à perdurer
    double moyenneLapinsVivants = 0;
    int nombreDisparitionEspece = 0;
    for(int i = 0; i < NBREXECUTIONSPOURSTATISTIQUEFIABLE; i++)
    {
        if(resultatsExecution[i].nbrLapinsVivants == 0)
        {
            nombreDisparitionEspece++;
        }
        moyenneLapinsVivants += resultatsExecution[i].nbrLapinsVivants;
        cout << "Execution n°" << (i + 1) << " au bout du " << NBRMOISEXPERIENECE << " ième mois : " << resultatsExecution[i].nbrLapinsVivants << " lapins vivants; " << resultatsExecution[i].nbrLapinsMorts << " lapins morts --> ça fait " << resultatsExecution[i].sommeNaissanceDeLapins << " naissances de lapins en tout.\n";
    }
    moyenneLapinsVivants /= NBREXECUTIONSPOURSTATISTIQUEFIABLE;
    float proportionPerdurationEspece = (1 - (float)((float)nombreDisparitionEspece / (float)NBREXECUTIONSPOURSTATISTIQUEFIABLE)) * 100;


    // Calcul du rayon de confiance à 95% pour pouvoir déterminer l'intervalle de confiance à 95% du nombre de lapins vivants à la fin d'une exécution avec les paramètres définis dans le general.hpp
    double  borneInfConfidenceInterval;
    double  borneSupConfidenceInterval;
    if(NBREXECUTIONSPOURSTATISTIQUEFIABLE - nombreDisparitionEspece >= 10)
    {
        double estimateurSansBiais = 0;
        for(int i = 0; i < NBREXECUTIONSPOURSTATISTIQUEFIABLE; i++)
        {
            if(resultatsExecution[i].nbrLapinsVivants > 0)
            {
                estimateurSansBiais += ((double)resultatsExecution[i].nbrLapinsVivants - moyenneLapinsVivants) * ((double)resultatsExecution[i].nbrLapinsVivants - moyenneLapinsVivants);
            }
        }
        estimateurSansBiais = estimateurSansBiais / (NBREXECUTIONSPOURSTATISTIQUEFIABLE - nombreDisparitionEspece - 1);
        
        int nTableOfT = NBREXECUTIONSPOURSTATISTIQUEFIABLE - nombreDisparitionEspece - 1;
        if(nTableOfT > 120){
            nTableOfT = 120;
        }
        double confidenceRadius = tableOfT[nTableOfT] * sqrt(estimateurSansBiais / (NBREXECUTIONSPOURSTATISTIQUEFIABLE - nombreDisparitionEspece));
        borneInfConfidenceInterval = moyenneLapinsVivants - confidenceRadius;
        borneSupConfidenceInterval = moyenneLapinsVivants + confidenceRadius;
    }


    cout << "\nIl y a donc eu en moyenne, sur " << NBREXECUTIONSPOURSTATISTIQUEFIABLE << " exécutions, " << moyenneLapinsVivants << " lapins vivants à la fin du " << NBRMOISEXPERIENECE << " ième mois.\n";
    cout << "L'espèce a perduré jusqu'à la fin de " << proportionPerdurationEspece << "% des exécutions\n\n";
    if(NBREXECUTIONSPOURSTATISTIQUEFIABLE - nombreDisparitionEspece >= 10)
    {
        cout << "L'espèce a perduré " << (NBREXECUTIONSPOURSTATISTIQUEFIABLE - nombreDisparitionEspece) << " fois, on peut donc réaliser une étude statistique :\n";
        cout << "Une fois que l'espèce a réussi à perdurer, le nombre de lapins vivants à la fin du " << NBRMOISEXPERIENECE << " ième mois a 95% de chances d'appartenir à l'intervalle [" << borneInfConfidenceInterval << " ; " << borneSupConfidenceInterval << "]\n";
    }
    else
    {
        cout << "La perduration de l'espèce est trop rare pour effectuer une étude statistique fiable\n";
    }

    cout << "Résultats obtenus avec NBRBEBESLAPINSINITIAUX = " << NBRBEBESLAPINSINITIAUX << " et PROBASEXEFEMININ = " << PROBASEXEFEMININ << "\n";
    cout << "Une femelle mature peut donc espérer obtenir " << ((MINPETITS + MAXPETITS) / 2.0) * PROBAPROCREATION << " petits par mois car MINPETITS = " << MINPETITS << "; MAXPETITS = " << MAXPETITS << "; PROBAPROCREATION = " << PROBAPROCREATION << ";\n"; 


    // On note la date de fin d'exécution.
	time_t end = time( NULL);

	// Impression de la date de compilation, la date de début d'exécution, la date de fin d'exécution, ainsi que le temps pris par l'exécution.
	std::string dateDebut = ctime(&begin);
    std::string dateFin = ctime(&end);

	std::cout << "\n\nCompilation effectuée le " << __DATE__ << " à " << __TIME__ << "\n";
    std::cout << "Exécution lancée le " << dateDebut;
    std::cout << "Exécution achevée le " << dateFin;

	unsigned long secondes = (unsigned long) difftime( end, begin );

    int jours = (int)(secondes / 86400);
    int secondesRestantes = (int)(secondes % 86400);
    int heures = secondesRestantes / 3600;
    secondesRestantes %= 3600;
    int minutes = secondesRestantes / 60;
    secondesRestantes %= 60;

    std::cout << "Le programme a donc été exécuté en " << secondes << " secondes : " << jours << " jours, " << heures << " heures, " << minutes << " minutes et " << secondesRestantes << " secondes\n";  
        
    return 0;
}